from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Union

import httpx
from pydantic import BaseModel


class BaseClient(ABC):
    def __init__(self, base_url: str, token: str):
        self.base_url = base_url
        self.headers = {
            "Authorization": f"Bearer {token}",
            "Content-Type": "application/json",
        }

    @abstractmethod
    async def _request(
        self,
        method: str,
        endpoint: str,
        *,
        json: Union[Dict[str, Any], List[Dict[str, Any]], None] = None,
        params: Optional[Union[Dict[str, Any], List[tuple[str, Any]]]] = None,
    ) -> Any:
        pass


class HttpClient(BaseClient):
    def __init__(self, base_url: str, token: str):
        super().__init__(base_url, token)
        self.client = httpx.AsyncClient()

    async def _request(
        self,
        method: str,
        endpoint: str,
        *,
        json: Union[Dict[str, Any], List[Dict[str, Any]], None] = None,
        params: Optional[Union[Dict[str, Any], List[tuple[str, Any]]]] = None,
        **kwargs: Dict[str, Any],
    ) -> Any:
        url = f"{self.base_url}{endpoint}"
        request_kwargs = {
            "method": method,
            "url": url,
            "headers": self.headers,
        }
        if json is not None:
            if isinstance(json, BaseModel):
                request_kwargs["json"] = json.model_dump(exclude_none=True)
            elif isinstance(json, (dict, list)):
                request_kwargs["json"] = json  # type: ignore
            else:
                raise ValueError(f"Unsupported JSON type: {type(json)}")
        for key, value in kwargs.items():
            if key not in request_kwargs:
                request_kwargs[key] = value

        async with self.client as client:
            response = await client.request(
                method=method,
                url=url,
                json=json,
                params=params,
                **kwargs,  # type: ignore
            )
            response.raise_for_status()
            return response.json()

    async def close(self) -> None:
        await self.client.aclose()



from typing import List, Optional, Union

from langchain_core.documents import Document
from langchain_openai import OpenAIEmbeddings
from langchain_text_splitters import CharacterTextSplitter, MarkdownTextSplitter

from whiskerrag_types.interface.embed_interface import BaseEmbedding
from whiskerrag_types.model.chunk import Chunk
from whiskerrag_types.model.knowledge import (
    EmbeddingModelEnum,
    Knowledge,
    KnowledgeTypeEnum,
)
from whiskerrag_utils import RegisterTypeEnum, register


@register(RegisterTypeEnum.EMBEDDING, EmbeddingModelEnum.OPENAI)
class OpenAIEmbedding(BaseEmbedding):
    async def embed_documents(
        self, knowledge: Knowledge, documents: List[Document]
    ) -> List[Chunk]:
        splitter: Optional[Union[CharacterTextSplitter, MarkdownTextSplitter]] = None
        if knowledge.knowledge_type == KnowledgeTypeEnum.TEXT:
            splitter = CharacterTextSplitter(
                chunk_size=knowledge.split_config.chunk_size,
                chunk_overlap=knowledge.split_config.chunk_overlap,
                separator=",".join(
                    x for x in (knowledge.split_config.separators or [])
                ),
            )
        if knowledge.knowledge_type == KnowledgeTypeEnum.MARKDOWN:
            splitter = MarkdownTextSplitter(
                chunk_size=knowledge.split_config.chunk_size,
                chunk_overlap=knowledge.split_config.chunk_overlap,
            )
        if splitter is None:
            raise Exception("not support knowledge type")

        chunks: List[Chunk] = []
        docs = splitter.split_documents(documents)
        embeddings = OpenAIEmbeddings()
        # TODO: rate limit
        for doc in docs:
            embedding = embeddings.embed_query(doc.page_content)
            chunk = Chunk(
                context=doc.page_content,
                metadata={
                    **doc.metadata,
                },
                embedding=embedding,
                knowledge_id=knowledge.knowledge_id,
                embedding_model_name=knowledge.embedding_model_name,
                space_id=knowledge.space_id,
            )
            chunks.append(chunk)
        return chunks

    async def embed_text(self, text: str) -> List[float]:
        embedding_client = OpenAIEmbeddings()
        embeddings = embedding_client.embed_query(text)
        return embeddings



from whiskerrag_client.chunk_client import ChunkClient
from whiskerrag_client.http_client import HttpClient
from whiskerrag_client.knowledge_client import KnowledgeClient
from whiskerrag_client.retrieval_client import RetrievalClient
from whiskerrag_client.task_client import TaskClient


class APIClient:
    def __init__(self, base_url: str, token: str):
        self.http_client = HttpClient(base_url, token)
        self.knowledge = KnowledgeClient(self.http_client)
        self.retrieval = RetrievalClient(self.http_client)
        self.chunk = ChunkClient(self.http_client)
        self.task = TaskClient(self.http_client)


__all__ = ["APIClient"]



from typing import Any, Dict, List, Optional

from whiskerrag_client.http_client import BaseClient
from whiskerrag_types.model.knowledge import Knowledge, KnowledgeCreate
from whiskerrag_types.model.page import PageParams, PageResponse


class KnowledgeClient:
    def __init__(self, http_client: BaseClient, base_path: str = "/api/knowledge"):
        self.http_client = http_client
        self.base_path = base_path

    async def add_knowledge(self, items: List[KnowledgeCreate]) -> List[Knowledge]:
        response = await self.http_client._request(
            method="POST",
            endpoint=f"{self.base_path}/add",
            json=[item.model_dump() for item in items],
        )
        return [Knowledge(**item) for item in response["data"]]

    async def get_knowledge_list(
        self,
        page: int = 1,
        page_size: int = 10,
        order_by: Optional[str] = None,
        order_direction: str = "asc",
        eq_conditions: Optional[Dict[str, Any]] = None,
    ) -> PageResponse[Knowledge]:
        params: PageParams = PageParams(
            page=page,
            page_size=page_size,
            order_by=order_by,
            order_direction=order_direction,
            eq_conditions=eq_conditions,
        )
        response = await self.http_client._request(
            method="POST",
            endpoint=f"{self.base_path}/list",
            json=params.model_dump(exclude_none=True),
        )
        return PageResponse(
            items=[Knowledge(**item) for item in response["data"]["items"]],
            total=response["data"]["total"],
            page=response["data"]["page"],
            page_size=response["data"]["page_size"],
            total_pages=response["data"]["total_page"],
        )

    async def get_knowledge_by_id(self, knowledge_id: str) -> Knowledge:
        response = await self.http_client._request(
            method="GET",
            endpoint=f"{self.base_path}/detail",
            params={"knowledge_id": knowledge_id},
        )
        return Knowledge(**response["data"])



from typing import List

from whiskerrag_client.http_client import BaseClient
from whiskerrag_types.model.page import PageResponse
from whiskerrag_types.model.retrieval import (
    RetrievalByKnowledgeRequest,
    RetrievalBySpaceRequest,
    RetrievalChunk,
)


class RetrievalClient:
    def __init__(self, http_client: BaseClient):
        self.http_client = http_client
        self.base_path = "/api/retrieval"

    async def retrieve_knowledge_content(
        self, request: RetrievalByKnowledgeRequest
    ) -> List[RetrievalChunk]:
        response = await self.http_client._request(
            method="POST", endpoint=f"{self.base_path}/knowledge", json=request.dict()
        )
        return [RetrievalChunk(**chunk) for chunk in response["data"]]

    async def retrieve_space_content(
        self, request: RetrievalBySpaceRequest
    ) -> PageResponse[RetrievalChunk]:
        response = await self.http_client._request(
            method="POST", endpoint=f"{self.base_path}/space", json=request.dict()
        )
        return PageResponse(
            items=[RetrievalChunk(**chunk) for chunk in response["data"]["items"]],
            total=response["data"]["total"],
            page=response["data"]["page"],
            page_size=response["data"]["page_size"],
            total_pages=response["data"]["totalPages"],
        )



from datetime import datetime
from typing import Any, List, Optional, Union
from uuid import uuid4

from pydantic import BaseModel, Field, field_serializer, field_validator

from whiskerrag_types.model.knowledge import EmbeddingModelEnum


class Chunk(BaseModel):
    chunk_id: str = Field(default_factory=lambda: str(uuid4()), description="chunk id")
    embedding: Optional[list[float]] = Field(None, description="chunk embedding")
    context: str = Field(..., description="chunk content")
    knowledge_id: str = Field(..., description="file source info")
    embedding_model_name: Optional[EmbeddingModelEnum] = Field(
        EmbeddingModelEnum.OPENAI, description="name of the embedding model"
    )
    space_id: str = Field(..., description="space id")
    metadata: Optional[dict] = Field(None, description="metadata")
    created_at: Optional[datetime] = Field(
        default_factory=lambda: datetime.now(), description="creation time"
    )
    updated_at: Optional[datetime] = Field(
        default_factory=lambda: datetime.now(), description="update time"
    )

    @field_validator("embedding", mode="before")
    @classmethod
    def parse_embedding(cls, v: Union[str, List[float], None]) -> Optional[List[float]]:
        if v is None:
            return None

        if isinstance(v, list):
            return [float(x) for x in v]

        if isinstance(v, str):
            v = v.strip()
            try:
                import json

                return (
                    [float(x) for x in json.loads(v)]
                    if isinstance(json.loads(v), list)
                    else None
                )
            except json.JSONDecodeError:
                try:
                    if v.startswith("[") and v.endswith("]"):
                        v = v[1:-1]
                    return [float(x.strip()) for x in v.split(",") if x.strip()]
                except ValueError:
                    raise ValueError(f"Invalid embedding format: {v}")

        raise ValueError(f"Unsupported embedding type: {type(v)}")

    @field_serializer("created_at")
    def serialize_created_at(self, created_at: Optional[datetime]) -> Optional[str]:
        return created_at.isoformat() if created_at else None

    @field_serializer("updated_at")
    def serialize_updated_at(self, updated_at: Optional[datetime]) -> Optional[str]:
        return updated_at.isoformat() if updated_at else None

    @field_serializer("embedding_model_name")
    def serialize_embedding_model_name(
        self, embedding_model_name: Optional[EmbeddingModelEnum]
    ) -> Optional[str]:
        return embedding_model_name.value if embedding_model_name else None

    def update(self, **kwargs: Any) -> "Chunk":
        for key, value in kwargs.items():
            setattr(self, key, value)

        self.updated_at = datetime.now()
        return self
